/*---------------------------------------------------------------------------*/
/* Program to get/set some ACPI settings                                     */
/* Author : David Leemans   -   http://freeunix.dyndns.org:8088/             */
/* Last update : 15-07-2004                                                  */
/*---------------------------------------------------------------------------*/
/*
 * acpitool.cpp
 *
 * Copyright (C) 2004  David Leemans <davidleemans |AT| tiscali |DOT| be>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/*---------------------------------------------------------------------------*/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <dirent.h>
#include "acpitool.h"

using namespace std;

#define App_Version "acpitool v0.2.1"

int Kernel_24 = 0, Kernel_26 = 0, be_quiet = 1;

/*----------------------------------------------------------------------------------------------------------------------------*/

int Has_ACPI()
{
    ifstream file_in;
    char *filename;
    
    filename = "/proc/acpi";
    
    file_in.open(filename);
    if (!file_in)
    	return 0;
    else
    {
	file_in.close();
        return 1;
    }
}
    
int Print_ACPI_Info(int show_ac, int show_therm, int show_fan, int show_batteries, int show_empty, int show_version, int info_level, int quiet)
{
    be_quiet = quiet;
	
    if(show_version)
	Do_SysVersion_Info();
    if(show_batteries)
    	Do_Battery_Stuff(show_empty, info_level);
    if(show_ac)
	Do_AC_Info();
    if(show_therm)
    	Do_Thermal_Info();
    if(show_fan)
    	Do_Fan_Info();
    	     
    return 0;
}

// remaining battery capacity (in %) = (remaining capacity / last full battery capacity) x 100 //
// remaining battery life (in hh:mm) = (remaining capacity / present battery rate ) //
   
int Do_Battery_Stuff(int show_empty, int info_level)
{
    Battery_Info *Batt_Info;
    int Bat_Nr;
    float Remaining_Percentage, Remaining_Time;
    int Time_In_Seconds, Hours, Minutes, Show_Time, Precision;
    
    int Nr_Batteries = Count_Batteries();
    if(Nr_Batteries>0)
    {
        Batt_Info = new Battery_Info;
		
        for(int i=0; i<Nr_Batteries; i++)
    	{
	    Bat_Nr = i+1;
	    		
	    Batt_Info->Battery_Present = 0;
	    strcpy(Batt_Info->Remaining_Cap,"         ");
	    strcpy(Batt_Info->LastFull_Cap,"         ");
	    strcpy(Batt_Info->Present_Rate,"         ");
	    strcpy(Batt_Info->Charging_State,"           ");
	    strcpy(Batt_Info->Technology, "            ");
	    strcpy(Batt_Info->Model, "            ");
	    strcpy(Batt_Info->Serial, "            ");
	    strcpy(Batt_Info->Bat_Type, "            ");     // initialize all struct members to blanks --> avoid rubbish in output //
			
	    Get_Battery_Info(Bat_Nr, Batt_Info);
	    Show_Time = atoi(Batt_Info->Present_Rate);       // avoid division by 0 if system is on AC power and battery is full and thus not charging //
	
	    switch(Batt_Info->Battery_Present)
	    {
		case 1 : 
		{	
	    	    Remaining_Percentage = float(atoi(Batt_Info->Remaining_Cap)) / float(atoi(Batt_Info->LastFull_Cap)) * 100.0;
            
            	    if( int(Remaining_Percentage) < 10)
            		Precision = 3;
            	    else
            		Precision = 4;
            	    
	    	    if(Show_Time)      // calculate remaining time only if present battery rate != 0 //
	    	    {
			    Remaining_Time = float(atoi(Batt_Info->Remaining_Cap)) / float(atoi(Batt_Info->Present_Rate)); 
			    // this represents hours //
		
			    Time_In_Seconds = int(Remaining_Time * 3600.0);
			    Hours = Time_In_Seconds / 3600;
			    Time_In_Seconds = Time_In_Seconds - (Hours * 3600);
			    Minutes = Time_In_Seconds / 60;
			    Time_In_Seconds = Time_In_Seconds - (Minutes * 60);
	    	    }
	    
	    	    switch(info_level)
	    	    {
			case 0 : cout.precision(Precision);
			     cout<<" Battery #"<<Bat_Nr<<"     : "<<Batt_Info->Charging_State<<", "<<showpoint<<Remaining_Percentage<<"%";
			     if(Show_Time)
			        cout<<", "<<setfill('0')<<setw(2)<<Hours<<":"<<setfill('0')<<setw(2)<<Minutes<<":"\
				<<setfill('0')<<setw(2)<<Time_In_Seconds;
			     cout<<endl; 
			     break;
			case 1 : cout<<" Battery #"<<Bat_Nr<<"     : present"<<endl;
			     cout<<"   Remaining capacity : "<<Batt_Info->Remaining_Cap<<", ";
			     cout.precision(Precision);
			     cout<<showpoint<<Remaining_Percentage<<"%";
			     if(Show_Time)
			     cout<<", "<<setfill('0')<<setw(2)<<Hours<<":"<<setfill('0')<<setw(2)<<Minutes<<":"\
			     <<setfill('0')<<setw(2)<<Time_In_Seconds;
			     cout<<endl;
			     cout<<"   Present rate       : "<<Batt_Info->Present_Rate<<endl;
			     cout<<"   Charging state     : "<<Batt_Info->Charging_State<<endl;
			     cout<<"   Last full capacity : "<<Batt_Info->LastFull_Cap<<endl;
			     cout<<"   Battery type       : "<<Batt_Info->Technology<<", "<<Batt_Info->Bat_Type<<endl;
			     cout<<"   Model number       : "<<Batt_Info->Model<<endl;
			     cout<<"   Serial number      : "<<Batt_Info->Serial<<endl;
			     break;
			default : cout<<"Undefined info level, fix this. Info level : "<<info_level<<endl;
			      break;
	    	    }
	    	    break;	
		}
	    case 0 : if(show_empty) cout<<" Battery #"<<Bat_Nr<<"     : slot empty"<<endl;
		     break;
	    case 2 : if(be_quiet)
			cout<<" Battery         : <not available>"<<endl;
		     else
			cout<<" Battery         : error reading info"<<endl;
		     break;
	    case 3 : cout<<" Battery         : <not available>"<<endl;
		     break;
	    default : cout<<"Undefined present value, fix this. Value : "<<Batt_Info->Battery_Present<<endl;
	              break;
	    }    
	
	}
	delete Batt_Info;
    }
    else
    {
	if(Nr_Batteries==0)   //this would apply to a regular desktop with acpi support //
	{
	    cout<<" Battery status : <not available>"<<endl;
	    return 0;
	}
	if(Nr_Batteries<0)
	{
	    if(be_quiet)
	    {
		cout<<" Battery status : <error reading info>"<<endl;
		return 0;
	    }
	    else
	    {
		cout<<" Battery status : <error reading info>"<<endl;
		cout<<" Function Count_Batteries returned : "<<Nr_Batteries<<endl;
		return -1;
	    }
	}
    }
    return 0;
}


int Get_ACPI_Version(char *c)
{
    ifstream file_in;
    char *filename, str[50];
    
    filename = "/proc/acpi/info";
    
    file_in.open(filename);
    if (!file_in)
    {
    	if(be_quiet)
    	{
	    strcpy(c, "<n.a>");
    	    return 0;
    	}
    	else
    	{
    	    cout<<" Could not open file : "<<filename<<endl;
	    cout<<" Make sure your kernel has ACPI support enabled"<<endl;
	    return -1;
    	}	
    }
       
    file_in.getline(str, 50);
    file_in.close();
    
    /* version takes 8 chars starting from pos 26 */
    strncpy(c, str+25,8);
    
    return 0;
}


int Get_Kernel_Version(char *c)
{
    ifstream file_in;
    char *filename, str[20];
    
    filename = "/proc/sys/kernel/osrelease";
    
    file_in.open(filename);
    if (!file_in)
    {
    	if(be_quiet)
    	{
    	    strcpy(c, "<n.a>");
    	    return 0;
    	}
    	else
    	{
    	    cout<<" Could not open file : "<<filename<<endl;
	    cout<<" Make sure your kernel has ACPI support enabled"<<endl;
	    return -1;
    	}	
    }
       
    file_in.getline(str, 16);
    file_in.close();
    
    strcpy(c, str);
    
    return 0;
}


int Do_SysVersion_Info()
{
    char Acpi_Version[10], Kernel_Version[15];
 
    strcpy(Acpi_Version,  "         "); 
    strcpy(Kernel_Version,  "              ");   

    Get_Kernel_Version(Kernel_Version);
    Get_ACPI_Version(Acpi_Version);

    cout<<" Kernel version : "<<Kernel_Version<<"  -  ACPI version : "<<Acpi_Version<<endl;
    cout<<" -------------------------------------------------------"<<endl;
    
    return 0;
}  


int Set_Kernel_Version()
{
    ifstream file_in;
    char *filename, str[10];
    
    filename = "/proc/sys/kernel/osrelease";
    
    file_in.open(filename);
    if (!file_in)
    {
    	cout<<" Could not open file : "<<filename<<endl;
	cout<<" function Set_Kernel_Version() : can't set kernel version, bailing out"<<endl;
	return -1;
    }	
       
    file_in.getline(str, 9);
    file_in.close();
    
    if(strncmp(str,"2.4",3)==0)
    {
	Kernel_24 = 1;           
	Kernel_26 = 0;
    }
        
    if(strncmp(str,"2.6",3)==0)
    {
	Kernel_24 = 0;           
	Kernel_26 = 1;
    }    
    
    return 0;
}


int Do_Thermal_Info()
{
    ifstream file_in;
    char *dirname, filename[4][50], str[40];
    int tz_count = 0;
    DIR *thermal_dir;
    struct dirent *therm;
    char *name;
    char Temperature[5], State[5];
 
    strcpy(Temperature, "    ");     
    strcpy(State, "    ");
    
    dirname = "/proc/acpi/thermal_zone/";   
    thermal_dir = opendir(dirname);
    if(thermal_dir)                    
    {
	struct dirent **namelist;
	int n;

	n = scandir(dirname, &namelist, 0, alphasort);
	if(n<0)
	    perror("scandir");
	else
	{
	    while(n--)
	    {
		name = namelist[n]->d_name;
		// skip . and .. //
		if (!strncmp (".", name, 1) || !strncmp ("..", name, 2)) continue;
		   		
		sprintf(filename[tz_count], "/proc/acpi/thermal_zone/%s/state", name);
       		tz_count++;
       		sprintf(filename[tz_count], "/proc/acpi/thermal_zone/%s/temperature", name);
    		tz_count++;
		free(namelist[n]);
	    }
	    free(namelist);
	}
    }      	
    else        
    {
       	if(be_quiet)    
    	{
	    cout<<" Thermal info   : <error>"<<endl;
      	    return 0;
    	}
    	else            
    	{
    	    cout<<" Could not open directory : "<<dirname<<endl;
    	    cout<<" function Do_Thermal_Info() : make sure your kernel has /proc/acpi/thermal_zone support enabled."<<endl;
	    return -1;
    	}	
    }
    closedir(thermal_dir);  
    
    if(tz_count>0)     
    {
    	tz_count--;
    	for(int i=tz_count; i>0; i-=2)     //process entries in reverse order //
    	{
    	    file_in.open(filename[i-1]);
    	    if(file_in)                   
    	    {
    		file_in.getline(str, 40);
    		strncpy(State, str+25, 5);
    		file_in.close();
    		cout<<" Thermal info   : "<<State<<", ";
    	    }
	    file_in.open(filename[i]);
    	    if(file_in)                   
    	    {
		file_in.getline(str, 40);
    		file_in.close();
    		strncpy(Temperature, str+25, 5);
    		cout<<Temperature<<endl;
	    }
	}
    }
    else   
	cout<<" Thermal info   : <not available>"<<endl;
    return 0;
}


int Do_AC_Info()
{
    ifstream file_in;
    char *dirname, filename[4][50], str[40];
    int ac_count = 0;
    DIR *ac_dir;
    struct dirent *ac_dirent;
    char *name;
    char AC_Status[9];
 
    strcpy(AC_Status, "        ");         // avoid rubbish in output //
      
    dirname = "/proc/acpi/ac_adapter/";    
    ac_dir = opendir(dirname);
    if(ac_dir)                             
    {
    	while((ac_dirent = readdir(ac_dir)))   
  	{
	    name = ac_dirent->d_name;
      	    // skip . and .. // 
    	    if (!strncmp (".", name, 1) || !strncmp ("..", name, 2)) continue;
       	    sprintf(filename[ac_count], "/proc/acpi/ac_adapter/%s/state", name);
       	    ac_count++;
       	}
    }
    else        
    {
       	if(be_quiet)   
    	{
    	    cout<<" AC adapter     : <error>"<<endl;
      	    return 0;
    	}
    	else              
    	{
    	    cout<<" Could not open directory : "<<dirname<<endl;
    	    cout<<" function Do_AC_Info() : make sure your kernel has /proc/acpi/ac_adapter support enabled."<<endl;
	    return -1;
    	}	
    }
    closedir(ac_dir);  
    
    if(ac_count>0)
    {
    	for(int i=0; i<ac_count; i++)
    	{
    	    file_in.open(filename[i]);
    	    if(file_in)                   
    	    {
		file_in.getline(str, 40);
    		strncpy(AC_Status, str+25, 8);
    		file_in.close();
    		cout<<" AC adapter     : "<<AC_Status<<endl;
    	    }
    	}
    }
    else
    	cout<<" AC adapter     : <not available>"<<endl;
    return 0;
}


int Get_Fan_Info(char *stat)
{
    ifstream file_in;
    char *filename, str[40];
    
    filename = "/proc/acpi/fan/FAN/state";
    
    file_in.open(filename);
    if (!file_in)        
    {
    	if(be_quiet)        
    	{
	    strcpy(stat, "<not available>");
    	    return 0;
    	}
    	else              
    	{
    	    cout<<" Could not open file : "<<filename<<endl;
	    cout<<" function Get_Fan_Info : make sure your kernel has /proc/acpi/fan support enabled."<<endl;
	    return -1;
    	}	
    }
           
    file_in.getline(str, 40);
    file_in.close();
    
    strncpy(stat, str+25, 4);
    
    return 0;
}


int Do_Fan_Info()
{
    char Fan_Status[16];
 
    strcpy(Fan_Status,  "               ");
  
    Get_Fan_Info(Fan_Status);
    
    cout<<" Fan            : "<<Fan_Status<<endl;
}


int Get_Battery_Info(const int bat_nr, Battery_Info *bat_info)
{
    ifstream file_in;
    char *dirname, filename[4][50], str[100], temp[5];
    int bat_count = 0, start = 0;
    DIR *battery_dir;
    struct dirent *bat_entry;
    char *name;
       
    dirname = "/proc/acpi/battery/";    //find all entries in this dir //
    battery_dir = opendir(dirname);
    if(battery_dir)                     // we can read this dir //
    {
	struct dirent **namelist;
	int n;

	n = scandir(dirname, &namelist, 0, alphasort);
	if(n<0)
	    perror("scandir");
	else
	{
	    while(n--)
	    {
		name = namelist[n]->d_name;
		// skip . and .. //
		if (!strncmp (".", name, 1) || !strncmp ("..", name, 2)) continue;
		
		sprintf(filename[bat_count], "/proc/acpi/battery/%s/info", name);
		bat_count++;
		sprintf(filename[bat_count], "/proc/acpi/battery/%s/state", name);
		bat_count++;
		free(namelist[n]);
	    }
	    free(namelist);
	}
    }      	
    else
    {
	if(be_quiet)
   	{
   	    bat_info->Battery_Present = 2;     // 2 represents error value //
   	    return 0;
   	}
   	else
   	{
   	    cout<<" Could not open directory : "<<dirname<<endl;
   	    cout<<" Make sure your kernel has /proc/acpi/battery/ support enabled"<<endl;
   	    return -1;
   	}
    }
    closedir(battery_dir);  
    
    //we found all dir entries, now process them //
    switch(bat_nr)       // select battery first //
    {
    	case 1 : start = 2;
    		 break;
    	case 2 : start = 0;
    		 break;      
    	default : cout<<"Invalid battery number, fix this"<<endl;
    		  return -1;
    		  break;
    } //NOTE : scandir returns entries in reverse order //
  
    // first get battery presence from 1st file//
    
    if(bat_count>0)
    {
    	file_in.open(filename[start]);
    	if (!file_in)
    	{
	    cout<<" Could not open file : "<<filename[start]<<endl;
	    cout<<" Make sure your kernel has /proc/acpi support enabled or check presence of a battery."<<endl;
	    return -1;
    	}
    	
    	file_in.getline(str, 100);
    	strncpy(temp, str+25, 4);
    	if(strncmp(temp,"yes",3)==0)
    	    bat_info->Battery_Present = 1;               //yes, we have a battery //
    	else
    	{
    	    bat_info->Battery_Present = 0;
    	    return 0;                  //bail out if battery is not present //
    	}
    	
    	// then get the last full capacity //
    	file_in.getline(str, 100);            //skip 1 line //
    	file_in.getline(str, 100);
    	strncpy(bat_info->LastFull_Cap, str+25, 9);
    
    	// then get the technology //
    	file_in.getline(str, 100);
    	strncpy(bat_info->Technology, str+25, 12);
    
    	// then get the model number //
    	for(int t=0; t<5; t++)
	file_in.getline(str, 100);            //skip 5 lines //
    	file_in.getline(str, 100);
    	strncpy(bat_info->Model, str+25, 12);
    
    	// then get the serial number //
    	file_in.getline(str, 100);
    	strncpy(bat_info->Serial, str+25, 12);
    
    	// then get the battery type //
    	file_in.getline(str, 100);
    	strncpy(bat_info->Bat_Type, str+25, 12);
    
    	file_in.close();
    	
    	// then open 2nd file = /proc/acpi/.../state //
    	file_in.open(filename[start+1]);
    	if (!file_in)
    	{
	    cout<<" Could not open file : "<<filename<<endl;
	    cout<<" Make sure your kernel has /proc/acpi support enabled"<<endl;
	    return -1;
    	}	

    	// then get the charging state //
    	file_in.getline(str, 100); file_in.getline(str, 100);     // skip first 2 lines //
    	file_in.getline(str, 100);
    	strncpy(bat_info->Charging_State, str+25, 12);

    	// then get the charging rate //
    	file_in.getline(str, 100);
    	strncpy(bat_info->Present_Rate, str+25, 9);

    	// then get the remaining capacity //
    	file_in.getline(str, 100);
    	strncpy(bat_info->Remaining_Cap, str+25, 9);
    
    	file_in.close();
	}
	else      // battery dir is readable but empty : only . and .. at most //
	    bat_info->Battery_Present = 3;   
	
	return 0;
}


int Count_Batteries()
{
    DIR *battery_dir;
    struct dirent *bat_entry;
    char *name, *dirname;
    int t = 0;
       
    dirname = "/proc/acpi/battery/";    //find all entries in this dir //
    battery_dir = opendir(dirname);
    if(battery_dir)                     // we can read this dir //
    {
	struct dirent **namelist;
	int n;

	n = scandir(dirname, &namelist, 0, alphasort);
	if(n<0)
	  return -1;
	else
	{
	    while(n--)
	    {
		name = namelist[n]->d_name;
		// skip . and .. //
		if (!strncmp (".", name, 1) || !strncmp ("..", name, 2))
		    continue;
		else
		    t++;
		free(namelist[n]);
	    }
	    free(namelist);
	    return t;
	}
    }
    else
	return 0;
}


int Do_Suspend(int State)                     //put system in sleep state 3 or 4 (standby) //
{
    ofstream file_out;
    char *filename, *str, ch;
    
    Set_Kernel_Version();
    
    switch(State)
    {
    	case 3 : if(Kernel_24)
		 {
		    filename = "/proc/acpi/sleep";
		    str = "3";
		 }
		 if(Kernel_26)
		 {
		    filename = "/sys/power/state";
		    str = "mem";
		 }
		 break;
	case 4 : if(Kernel_24)
		 {
		    filename = "/proc/acpi/sleep";
		    str = "4";
		 }
		 if(Kernel_26)
		 {
		    filename = "/sys/power/state";
		    str = "disk";
		 }
		 break;
	default : cout<<"Unsupported state, fix this"<<endl;
		  break;
    }
    
    file_out.open(filename);
    if (!file_out)
    {
    	cout<<" Could not open file : "<<filename<<". \n"
	      " You must have write access to "<<filename<<" to suspend your computer."<<endl;
	return -1;
    }	
       
    file_out<<str;     // sleep tight //
    
    return 0;
}


int Show_Help()
{
    cout<<"Usage: acpitool [option] . . . \n"
"Shows ACPI information from the /proc/acpi filesystem, like battery status,\n"
"temperature, or ac power. Can also suspend your machine (if supported).\n"
"\n"
"  -a      AC adapter information\n"
"  -b      battery status, available batteries only\n"
"  -B      battery status, all batteries\n"
"  -e      show extended information\n"
"  -f      show fan status\n"
"  -F x    force fan on (x=1) or off(x=0). (Toshiba only)\n"
"  -h      show this help screen\n"
"  -l x    set LCD brightness level to x, where x is 0..7 (Toshiba only)\n"
"  -s      suspend (sleep state S3), if supported\n"
"  -S      suspend (sleep state S4), if supported\n"
"  -t      thermal information\n"
"  -T      show supported Toshiba ACPI extensions (LCD brightness level, video out routing, fan status)\n"
"  -v      show application version number\n"
"\n"
"If invoked without options, acpitool displays information about available batteries,\n"
"AC adapter and thermal information.\n"
"Essentially, this is the same as invoking acpitool with the -bat options."
"\n";
}


int Show_App_Version()
{
    cout<<App_Version<<endl;
    cout<<"Copyright (C) 2004 David Leemans.\n"
     "This program is free software; you can redistribute it and/or modify\n"
     "it under the terms of the GNU General Public License as published by\n"
     "the Free Software Foundation; either version 2 of the License, or\n"
     "(at your option) any later version.\n"
     "\n"
     "This program is distributed in the hope that it will be useful,\n"
     "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
     "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
     "GNU General Public License for more details.\n";
}
